from __future__ import annotations

from typing import List

from kr8s.asyncio.objects import Deployment, Pod, Service, new_class


class DaskCluster(new_class("DaskCluster", "kubernetes.dask.org/v1")):
    scalable = True
    scalable_spec = "worker.replicas"

    async def worker_groups(self) -> List[DaskWorkerGroup]:
        return [
            wg
            async for wg in DaskWorkerGroup.list(
                label_selector=f"dask.org/cluster-name={self.name}",
                namespace=self.namespace,
            )
        ]

    async def scheduler_pod(self) -> Pod:
        pods = []
        while not pods:
            pods = [
                pod
                async for pod in Pod.list(
                    label_selector=",".join(
                        [
                            f"dask.org/cluster-name={self.name}",
                            "dask.org/component=scheduler",
                        ]
                    ),
                    namespace=self.namespace,
                )
            ]
        assert len(pods) == 1
        return pods[0]

    async def scheduler_deployment(self) -> Deployment:
        deployments = []
        while not deployments:
            deployments = [
                deployment
                async for deployment in Deployment.list(
                    label_selector=",".join(
                        [
                            f"dask.org/cluster-name={self.name}",
                            "dask.org/component=scheduler",
                        ]
                    ),
                    namespace=self.namespace,
                )
            ]
        assert len(deployments) == 1
        return deployments[0]

    async def scheduler_service(self) -> Service:
        services = []
        while not services:
            services = [
                service
                async for service in Service.list(
                    label_selector=",".join(
                        [
                            f"dask.org/cluster-name={self.name}",
                            "dask.org/component=scheduler",
                        ]
                    ),
                    namespace=self.namespace,
                )
            ]
        assert len(services) == 1
        return services[0]

    async def ready(self) -> bool:
        await self.async_refresh()
        return (
            "status" in self.raw
            and "phase" in self.status
            and self.status.phase == "Running"
        )


class DaskWorkerGroup(new_class("DaskWorkerGroup", "kubernetes.dask.org/v1")):
    scalable = True
    scalable_spec = "worker.replicas"

    async def pods(self) -> List[Pod]:
        return [
            pod
            async for pod in Pod.list(
                label_selector=",".join(
                    [
                        f"dask.org/cluster-name={self.spec.cluster}",
                        "dask.org/component=worker",
                        f"dask.org/workergroup-name={self.name}",
                    ]
                ),
                namespace=self.namespace,
            )
        ]

    async def deployments(self) -> List[Deployment]:
        return [
            deployment
            async for deployment in Deployment.list(
                label_selector=",".join(
                    [
                        f"dask.org/cluster-name={self.spec.cluster}",
                        "dask.org/component=worker",
                        f"dask.org/workergroup-name={self.name}",
                    ]
                ),
                namespace=self.namespace,
            )
        ]

    async def cluster(self) -> DaskCluster:
        return await DaskCluster.get(self.spec.cluster, namespace=self.namespace)


class DaskAutoscaler(new_class("DaskAutoscaler", "kubernetes.dask.org/v1")):
    async def cluster(self) -> DaskCluster:
        return await DaskCluster.get(self.spec.cluster, namespace=self.namespace)


class DaskJob(new_class("DaskJob", "kubernetes.dask.org/v1")):
    async def cluster(self) -> DaskCluster:
        return await DaskCluster.get(self.name, namespace=self.namespace)

    async def pod(self) -> Pod:
        pods = []
        while not pods:
            pods = [
                pod
                async for pod in Pod.list(
                    label_selector=",".join(
                        [
                            f"dask.org/cluster-name={self.name}",
                            "dask.org/component=job-runner",
                        ]
                    ),
                    namespace=self.namespace,
                )
            ]
        assert len(pods) == 1
        return pods[0]



from __future__ import annotations

import asyncio
import json
import os.path
import pathlib
from contextlib import asynccontextmanager
from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    AsyncIterator,
    Callable,
    Final,
    Iterator,
)

import dask.config
import pytest
import yaml
from dask.distributed import Client
from kr8s.asyncio.objects import Deployment, Pod, Service

from dask_kubernetes.constants import MAX_CLUSTER_NAME_LEN
from dask_kubernetes.operator._objects import DaskCluster, DaskJob, DaskWorkerGroup
from dask_kubernetes.operator.controller import (
    KUBERNETES_DATETIME_FORMAT,
    get_job_runner_pod_name,
    retire_workers_lifo,
)

if TYPE_CHECKING:
    from kopf.testing import KopfRunner
    from pytest_kind.cluster import KindCluster

DIR: Final[pathlib.Path] = pathlib.Path(__file__).parent.absolute()

_EXPECTED_ANNOTATIONS: Final[dict[str, str]] = {"test-annotation": "annotation-value"}
_EXPECTED_LABELS: Final[dict[str, str]] = {"test-label": "label-value"}
DEFAULT_CLUSTER_NAME: Final[str] = "simple"


@pytest.fixture()
def gen_cluster_manifest(tmp_path: pathlib.Path) -> Callable[..., pathlib.Path]:
    def factory(cluster_name: str = DEFAULT_CLUSTER_NAME) -> pathlib.Path:
        original_manifest_path = os.path.join(DIR, "resources", "simplecluster.yaml")
        with open(original_manifest_path, "r") as original_manifest_file:
            manifest = yaml.safe_load(original_manifest_file)

        manifest["metadata"]["name"] = cluster_name
        new_manifest_path = tmp_path / "cluster.yaml"
        new_manifest_path.write_text(yaml.safe_dump(manifest))
        return tmp_path

    return factory


@pytest.fixture()
def gen_cluster(
    k8s_cluster: KindCluster,
    namespace: str,
    gen_cluster_manifest: Callable[..., pathlib.Path],
) -> Iterator[Callable[..., AsyncContextManager[tuple[str, str]]]]:
    """Yields an instantiated context manager for creating/deleting a simple cluster."""

    @asynccontextmanager
    async def cm(
        cluster_name: str = DEFAULT_CLUSTER_NAME,
    ) -> AsyncIterator[tuple[str, str]]:
        cluster_path = gen_cluster_manifest(cluster_name)
        # Create cluster resource
        k8s_cluster.kubectl("apply", "-n", namespace, "-f", str(cluster_path))
        while cluster_name not in k8s_cluster.kubectl(
            "get", "daskclusters.kubernetes.dask.org", "-n", namespace
        ):
            await asyncio.sleep(0.1)

        try:
            yield cluster_name, namespace
        finally:
            # Test: remove the wait=True, because I think this is blocking the operator
            k8s_cluster.kubectl("delete", "-n", namespace, "-f", str(cluster_path))

    yield cm


@pytest.fixture()
def gen_job(
    k8s_cluster: KindCluster, namespace: str
) -> Iterator[Callable[[str], AsyncContextManager[tuple[str, str]]]]:
    """Yields an instantiated context manager for creating/deleting a simple job."""

    @asynccontextmanager
    async def cm(job_file: str) -> AsyncIterator[tuple[str, str]]:
        job_path = os.path.join(DIR, "resources", job_file)
        with open(job_path) as f:
            job_name = yaml.load(f, yaml.Loader)["metadata"]["name"]

        # Create cluster resource
        k8s_cluster.kubectl("apply", "-n", namespace, "-f", job_path)
        while job_name not in k8s_cluster.kubectl(
            "get", "daskjobs.kubernetes.dask.org", "-n", namespace
        ):
            await asyncio.sleep(0.1)

        try:
            yield job_name, namespace
        finally:
            # Test: remove the wait=True, because I think this is blocking the operator
            k8s_cluster.kubectl("delete", "-n", namespace, "-f", job_path)
            while job_name in k8s_cluster.kubectl(
                "get", "daskjobs.kubernetes.dask.org", "-n", namespace
            ):
                await asyncio.sleep(0.1)

    yield cm


@pytest.fixture()
def gen_worker_group(
    k8s_cluster: KindCluster, namespace: str
) -> Iterator[Callable[[str], AsyncContextManager[tuple[str, str]]]]:
    """Yields an instantiated context manager for creating/deleting a worker group."""

    @asynccontextmanager
    async def cm(worker_group_file: str) -> AsyncIterator[tuple[str, str]]:
        worker_group_path = os.path.join(DIR, "resources", worker_group_file)
        with open(worker_group_path) as f:
            worker_group_name = yaml.load(f, yaml.Loader)["metadata"]["name"]

        # Create cluster resource
        k8s_cluster.kubectl("apply", "-n", namespace, "-f", worker_group_path)
        while worker_group_name not in k8s_cluster.kubectl(
            "get", "daskworkergroups.kubernetes.dask.org", "-n", namespace
        ):
            await asyncio.sleep(0.1)

        try:
            yield worker_group_name, namespace
        finally:
            # Test: remove the wait=True, because I think this is blocking the operator
            k8s_cluster.kubectl("delete", "-n", namespace, "-f", worker_group_path)
            while worker_group_name in k8s_cluster.kubectl(
                "get", "daskworkergroups.kubernetes.dask.org", "-n", namespace
            ):
                await asyncio.sleep(0.1)

    yield cm


def test_customresources(k8s_cluster: KindCluster) -> None:
    assert "daskclusters.kubernetes.dask.org" in k8s_cluster.kubectl("get", "crd")
    assert "daskworkergroups.kubernetes.dask.org" in k8s_cluster.kubectl("get", "crd")
    assert "daskjobs.kubernetes.dask.org" in k8s_cluster.kubectl("get", "crd")


def test_operator_runs(kopf_runner: KopfRunner) -> None:
    with kopf_runner as runner:
        pass

    assert runner.exit_code == 0
    assert runner.exception is None


def test_operator_plugins(kopf_runner: KopfRunner) -> None:
    with kopf_runner as runner:
        pass

    assert runner.exit_code == 0
    assert runner.exception is None
    assert "Plugin 'noop' running." in runner.output


@pytest.mark.timeout(180)
@pytest.mark.anyio
async def test_simplecluster(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_cluster: Callable[..., AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        async with gen_cluster() as (cluster_name, ns):
            scheduler_deployment_name = "simple-scheduler"
            worker_pod_name = "simple-default-worker"
            service_name = "simple-scheduler"

            while scheduler_deployment_name not in k8s_cluster.kubectl(
                "get", "deployments", "-n", ns
            ):
                await asyncio.sleep(0.1)
            while service_name not in k8s_cluster.kubectl("get", "svc", "-n", ns):
                await asyncio.sleep(0.1)
            while worker_pod_name not in k8s_cluster.kubectl("get", "pods", "-n", ns):
                await asyncio.sleep(0.1)

            with k8s_cluster.port_forward(
                f"service/{service_name}", 8786, "-n", ns
            ) as port:
                async with Client(
                    f"tcp://localhost:{port}", asynchronous=True
                ) as client:
                    await client.wait_for_workers(2)
                    # Ensure that inter-worker communication works well
                    futures = client.map(lambda x: x + 1, range(10))
                    total = client.submit(sum, futures)
                    assert (await total) == sum(map(lambda x: x + 1, range(10)))

            # Get the first env value (the only one) of the scheduler
            scheduler_env = k8s_cluster.kubectl(
                "get",
                "pods",
                "-n",
                ns,
                "--selector=dask.org/component=scheduler",
                "-o",
                "jsonpath='{.items[0].spec.containers[0].env[0]}'",
            )
            # Just check if it's in the string, no need to parse the json
            assert "SCHEDULER_ENV" in scheduler_env

            # Get the first annotation (the only one) of the scheduler
            scheduler_annotations = json.loads(
                k8s_cluster.kubectl(
                    "get",
                    "pods",
                    "-n",
                    ns,
                    "--selector=dask.org/component=scheduler",
                    "-o",
                    "jsonpath='{.items[0].metadata.annotations}'",
                )[1:-1]
            )  # First and last char is a quote
            assert _EXPECTED_ANNOTATIONS.items() <= scheduler_annotations.items()

            # Get the first annotation (the only one) of the scheduler
            service_annotations = json.loads(
                k8s_cluster.kubectl(
                    "get",
                    "svc",
                    "-n",
                    ns,
                    "--selector=dask.org/cluster-name=simple",
                    "-o",
                    "jsonpath='{.items[0].metadata.annotations}'",
                )[1:-1]
            )  # First and last char is a quote
            assert _EXPECTED_ANNOTATIONS.items() <= service_annotations.items()

            # Get the first env value (the only one) of the first worker
            worker_env = k8s_cluster.kubectl(
                "get",
                "pods",
                "-n",
                ns,
                "--selector=dask.org/component=worker",
                "-o",
                "jsonpath='{.items[0].spec.containers[0].env[0]}'",
            )
            # Just check if it's in the string, no need to parse the json
            assert "WORKER_ENV" in worker_env
            assert cluster_name

            # Get the first annotation (the only one) of a worker
            worker_annotations = json.loads(
                k8s_cluster.kubectl(
                    "get",
                    "pods",
                    "-n",
                    ns,
                    "--selector=dask.org/component=worker",
                    "-o",
                    "jsonpath='{.items[0].metadata.annotations}'",
                )[1:-1]
            )
            assert _EXPECTED_ANNOTATIONS.items() <= worker_annotations.items()

            # Assert labels from the dask cluster are propagated into the dask worker group
            workergroup_labels = json.loads(
                k8s_cluster.kubectl(
                    "get",
                    "daskworkergroups",
                    "-n",
                    ns,
                    "--selector=dask.org/component=workergroup",
                    "-o",
                    "jsonpath='{.items[0].metadata.labels}'",
                )[1:-1]
            )
            assert _EXPECTED_LABELS.items() <= workergroup_labels.items()
            assert "worker-sublabel" in workergroup_labels

            # Assert labels from the dask cluster are propagated into the dask scheduler
            scheduler_labels = json.loads(
                k8s_cluster.kubectl(
                    "get",
                    "pods",
                    "-n",
                    ns,
                    "--selector=dask.org/component=scheduler",
                    "-o",
                    "jsonpath='{.items[0].metadata.labels}'",
                )[1:-1]
            )
            assert _EXPECTED_LABELS.items() <= scheduler_labels.items()
            assert "scheduler-sublabel" in scheduler_labels

            # Assert labels from the dask cluster are propagated into the dask worker pod
            worker_labels = json.loads(
                k8s_cluster.kubectl(
                    "get",
                    "pods",
                    "-n",
                    ns,
                    "--selector=dask.org/component=worker",
                    "-o",
                    "jsonpath='{.items[0].metadata.labels}'",
                )[1:-1]
            )
            assert _EXPECTED_LABELS.items() <= worker_labels.items()
            assert "worker-sublabel" in workergroup_labels


@pytest.mark.anyio
async def test_scalesimplecluster(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_cluster: Callable[..., AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        async with gen_cluster() as (cluster_name, ns):
            scheduler_deployment_name = "simple-scheduler"
            worker_pod_name = "simple-default-worker"
            service_name = "simple-scheduler"
            while scheduler_deployment_name not in k8s_cluster.kubectl(
                "get", "deployments", "-n", ns
            ):
                await asyncio.sleep(0.1)
            while service_name not in k8s_cluster.kubectl("get", "svc", "-n", ns):
                await asyncio.sleep(0.1)
            while worker_pod_name not in k8s_cluster.kubectl("get", "pods", "-n", ns):
                await asyncio.sleep(0.1)

            with k8s_cluster.port_forward(
                f"service/{service_name}", 8786, "-n", ns
            ) as port:
                async with Client(
                    f"tcp://localhost:{port}", asynchronous=True
                ) as client:
                    k8s_cluster.kubectl(
                        "scale",
                        "-n",
                        ns,
                        "--replicas=5",
                        "daskworkergroup.kubernetes.dask.org",
                        "simple-default",
                    )
                    await client.wait_for_workers(5)
                    k8s_cluster.kubectl(
                        "scale",
                        "-n",
                        ns,
                        "--replicas=3",
                        "daskworkergroup.kubernetes.dask.org",
                        "simple-default",
                    )
                    # TODO: Currently, doesn't test anything. Need to add optional
                    #       argument to wait when removing workers once distributed
                    #       PR github.com/dask/distributed/pull/6377 is merged.
                    await client.wait_for_workers(3)


@pytest.mark.anyio
async def test_scalesimplecluster_from_cluster_spec(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_cluster: Callable[..., AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        async with gen_cluster() as (cluster_name, ns):
            scheduler_deployment_name = "simple-scheduler"
            worker_pod_name = "simple-default-worker"
            service_name = "simple-scheduler"
            while scheduler_deployment_name not in k8s_cluster.kubectl(
                "get", "pods", "-n", ns
            ):
                await asyncio.sleep(0.1)
            while service_name not in k8s_cluster.kubectl("get", "svc", "-n", ns):
                await asyncio.sleep(0.1)
            while worker_pod_name not in k8s_cluster.kubectl("get", "pods", "-n", ns):
                await asyncio.sleep(0.1)

            with k8s_cluster.port_forward(
                f"service/{service_name}", 8786, "-n", ns
            ) as port:
                async with Client(
                    f"tcp://localhost:{port}", asynchronous=True
                ) as client:
                    k8s_cluster.kubectl(
                        "scale",
                        "-n",
                        ns,
                        "--replicas=5",
                        "daskcluster.kubernetes.dask.org",
                        cluster_name,
                    )
                    await client.wait_for_workers(5)
                    k8s_cluster.kubectl(
                        "scale",
                        "-n",
                        ns,
                        "--replicas=3",
                        "daskcluster.kubernetes.dask.org",
                        cluster_name,
                    )
                    # TODO: Currently, doesn't test anything. Need to add optional
                    #       argument to wait when removing workers once distributed
                    #       PR github.com/dask/distributed/pull/6377 is merged.
                    await client.wait_for_workers(3)


@pytest.mark.anyio
async def test_recreate_scheduler_pod(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_cluster: Callable[..., AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        async with gen_cluster() as (cluster_name, ns):
            scheduler_deployment_name = "simple-scheduler"
            worker_pod_name = "simple-default-worker"
            service_name = "simple-scheduler"
            while scheduler_deployment_name not in k8s_cluster.kubectl(
                "get", "pods", "-n", ns
            ):
                await asyncio.sleep(0.1)
            while service_name not in k8s_cluster.kubectl("get", "svc", "-n", ns):
                await asyncio.sleep(0.1)
            while worker_pod_name not in k8s_cluster.kubectl("get", "pods", "-n", ns):
                await asyncio.sleep(0.1)
            k8s_cluster.kubectl(
                "delete",
                "pods",
                "-l",
                "dask.org/cluster-name=simple,dask.org/component=scheduler",
                "-n",
                ns,
            )
            k8s_cluster.kubectl(
                "wait",
                "--for=condition=Ready",
                "-l",
                "dask.org/cluster-name=simple,dask.org/component=scheduler",
                "pod",
                "-n",
                ns,
                "--timeout=60s",
            )
            assert scheduler_deployment_name in k8s_cluster.kubectl(
                "get", "pods", "-n", ns
            )


@pytest.mark.anyio
@pytest.mark.skip(reason="Flaky in CI")
async def test_recreate_worker_pods(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_cluster: Callable[..., AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        async with gen_cluster() as (cluster_name, ns):
            cluster = await DaskCluster.get(cluster_name, namespace=ns)
            # Get the default worker group
            while not (wgs := await cluster.worker_groups()):
                await asyncio.sleep(0.1)
            [wg] = wgs
            # Wait for worker Pods to be created
            while not (pods := await wg.pods()):
                await asyncio.sleep(0.1)
            # Store number of workers
            n_pods = len(pods)
            # Wait for worker Pods to be ready
            await asyncio.gather(
                *[pod.wait(conditions="condition=Ready", timeout=60) for pod in pods]
            )
            # Delete a worker Pod
            await pods[0].delete()
            # Wait for Pods to be recreated
            while len((pods := await wg.pods())) < n_pods:
                await asyncio.sleep(0.1)
            # Wait for worker Pods to be ready
            await asyncio.gather(
                *[pod.wait(conditions="condition=Ready", timeout=60) for pod in pods]
            )


@pytest.mark.anyio
async def test_simplecluster_batched_worker_deployments(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_cluster: Callable[..., AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        with dask.config.set(
            {
                "kubernetes.controller.worker-allocation.batch-size": 1,
                "kubernetes.controller.worker-allocation.delay": 5,
            }
        ):
            async with gen_cluster() as (cluster_name, ns):
                scheduler_deployment_name = "simple-scheduler"
                worker_pod_name = "simple-default-worker"
                service_name = "simple-scheduler"
                while scheduler_deployment_name not in k8s_cluster.kubectl(
                    "get", "deployments", "-n", ns
                ):
                    await asyncio.sleep(0.1)
                while service_name not in k8s_cluster.kubectl("get", "svc", "-n", ns):
                    await asyncio.sleep(0.1)
                while worker_pod_name not in k8s_cluster.kubectl(
                    "get", "pods", "-n", ns
                ):
                    await asyncio.sleep(0.1)

                with k8s_cluster.port_forward(
                    f"service/{service_name}", 8786, "-n", ns
                ) as port:
                    async with Client(
                        f"tcp://localhost:{port}", asynchronous=True
                    ) as client:
                        await client.wait_for_workers(2)
                        futures = client.map(lambda x: x + 1, range(10))
                        total = client.submit(sum, futures)
                        assert (await total) == sum(map(lambda x: x + 1, range(10)))


def _get_job_status(k8s_cluster: KindCluster, ns: str) -> dict[str, Any]:
    return json.loads(  # type: ignore[no-any-return]
        k8s_cluster.kubectl(
            "get",
            "-n",
            ns,
            "daskjobs.kubernetes.dask.org",
            "-o",
            "jsonpath='{.items[0].status}'",
        )[1:-1]
    )


def _assert_job_status_created(job_status: dict[str, Any]) -> None:
    assert "jobStatus" in job_status


def _assert_job_status_cluster_created(job: str, job_status: dict[str, Any]) -> None:
    assert "jobStatus" in job_status
    assert job_status["clusterName"] == job
    assert job_status["jobRunnerPodName"] == get_job_runner_pod_name(job)


def _assert_job_status_running(job: str, job_status: dict[str, Any]) -> None:
    assert "jobStatus" in job_status
    assert job_status["clusterName"] == job
    assert job_status["jobRunnerPodName"] == get_job_runner_pod_name(job)
    start_time = datetime.strptime(job_status["startTime"], KUBERNETES_DATETIME_FORMAT)
    assert datetime.utcnow() > start_time > (datetime.utcnow() - timedelta(seconds=10))


def _assert_final_job_status(
    job: str, job_status: dict[str, Any], expected_status: str
) -> None:
    assert job_status["jobStatus"] == expected_status
    assert job_status["clusterName"] == job
    assert job_status["jobRunnerPodName"] == get_job_runner_pod_name(job)
    start_time = datetime.strptime(job_status["startTime"], KUBERNETES_DATETIME_FORMAT)
    assert datetime.utcnow() > start_time > (datetime.utcnow() - timedelta(minutes=1))
    end_time = datetime.strptime(job_status["endTime"], KUBERNETES_DATETIME_FORMAT)
    assert datetime.utcnow() > end_time > (datetime.utcnow() - timedelta(minutes=1))
    assert set(job_status.keys()) == {
        "clusterName",
        "jobRunnerPodName",
        "jobStatus",
        "startTime",
        "endTime",
    }


@pytest.mark.anyio
async def test_job(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_job: Callable[[str], AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner as runner:
        async with gen_job("simplejob.yaml") as (job, ns):
            assert job

            runner_name = f"{job}-runner"

            # Assert that job was created
            while job not in k8s_cluster.kubectl(
                "get", "daskjobs.kubernetes.dask.org", "-n", ns
            ):
                await asyncio.sleep(0.1)

            job_status = _get_job_status(k8s_cluster, ns)
            _assert_job_status_created(job_status)

            # Assert that cluster is created
            while job not in k8s_cluster.kubectl(
                "get", "daskclusters.kubernetes.dask.org", "-n", ns
            ):
                await asyncio.sleep(0.1)

            await asyncio.sleep(0.1)  # Wait for a short time, to avoid race condition
            job_status = _get_job_status(k8s_cluster, ns)
            _assert_job_status_cluster_created(job, job_status)

            # Assert job pod is created
            while job not in k8s_cluster.kubectl("get", "po", "-n", ns):
                await asyncio.sleep(0.1)

            # Assert that pod started Running
            while "Running" not in k8s_cluster.kubectl("get", "po", "-n", ns):
                await asyncio.sleep(0.1)

            await asyncio.sleep(5)  # Wait for a short time, to avoid race condition
            job_status = _get_job_status(k8s_cluster, ns)
            _assert_job_status_running(job, job_status)

            job_annotations = json.loads(
                k8s_cluster.kubectl(
                    "get",
                    "pods",
                    "-n",
                    ns,
                    "--selector=dask.org/component=job-runner",
                    "-o",
                    "jsonpath='{.items[0].metadata.annotations}'",
                )[1:-1]
            )
            assert _EXPECTED_ANNOTATIONS.items() <= job_annotations.items()

            # Assert job pod runs to completion (will fail if doesn't connect to cluster)
            while "Completed" not in k8s_cluster.kubectl(
                "get", "-n", ns, "po", runner_name
            ):
                await asyncio.sleep(0.1)

            # Assert cluster is removed on completion
            while job in k8s_cluster.kubectl(
                "get", "-n", ns, "daskclusters.kubernetes.dask.org"
            ):
                await asyncio.sleep(0.1)

            job_status = _get_job_status(k8s_cluster, ns)
            _assert_final_job_status(job, job_status, "Successful")

    assert "A DaskJob has been created" in runner.output
    assert "Job succeeded, deleting Dask cluster." in runner.output


@pytest.mark.anyio
async def test_failed_job(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_job: Callable[[str], AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner as runner:
        async with gen_job("failedjob.yaml") as (job, ns):
            assert job

            runner_name = f"{job}-runner"

            # Assert that job was created
            while job not in k8s_cluster.kubectl(
                "get", "daskjobs.kubernetes.dask.org", "-n", ns
            ):
                await asyncio.sleep(0.1)

            job_status = _get_job_status(k8s_cluster, ns)
            _assert_job_status_created(job_status)

            # Assert that cluster is created
            while job not in k8s_cluster.kubectl(
                "get", "daskclusters.kubernetes.dask.org", "-n", ns
            ):
                await asyncio.sleep(0.1)

            await asyncio.sleep(0.1)  # Wait for a short time, to avoid race condition
            job_status = _get_job_status(k8s_cluster, ns)
            _assert_job_status_cluster_created(job, job_status)

            # Assert job pod is created
            while job not in k8s_cluster.kubectl("get", "po", "-n", ns):
                await asyncio.sleep(0.1)

            # Assert that pod started Running
            while "Running" not in k8s_cluster.kubectl("get", "po", "-n", ns):
                await asyncio.sleep(0.1)

            await asyncio.sleep(5)  # Wait for a short time, to avoid race condition
            job_status = _get_job_status(k8s_cluster, ns)
            _assert_job_status_running(job, job_status)

            # Assert job pod runs to failure
            while "Error" not in k8s_cluster.kubectl(
                "get", "po", "-n", ns, runner_name
            ):
                await asyncio.sleep(0.1)

            # Assert cluster is removed on completion
            while job in k8s_cluster.kubectl(
                "get", "-n", ns, "daskclusters.kubernetes.dask.org"
            ):
                await asyncio.sleep(0.1)

            job_status = _get_job_status(k8s_cluster, ns)
            _assert_final_job_status(job, job_status, "Failed")

    assert "A DaskJob has been created" in runner.output
    assert "Job failed, deleting Dask cluster." in runner.output


@pytest.mark.anyio
async def test_object_dask_cluster(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_cluster: Callable[..., AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        async with gen_cluster() as (cluster_name, ns):
            cluster = await DaskCluster.get(cluster_name, namespace=ns)

            worker_groups: list[DaskWorkerGroup] = []
            while not worker_groups:
                worker_groups = await cluster.worker_groups()
                await asyncio.sleep(0.1)
            assert len(worker_groups) == 1  # Just the default worker group
            wg = worker_groups[0]
            assert isinstance(wg, DaskWorkerGroup)

            # Test for non-replicated environment variables; Fix for https://github.com/dask/dask-kubernetes/issues/841
            for deployment in await wg.deployments():
                env_vars = deployment.spec["template"]["spec"]["containers"]["env"]
                env_var_names = [env_var["name"] for env_var in env_vars]
                assert len(env_var_names) == len(set(env_var_names))
                assert "DASK_WORKER_NAME" in env_var_names
                assert "DASK_SCHEDULER_ADDRESS" in env_var_names

            scheduler_pod = await cluster.scheduler_pod()
            assert isinstance(scheduler_pod, Pod)

            scheduler_deployment = await cluster.scheduler_deployment()
            assert isinstance(scheduler_deployment, Deployment)

            scheduler_service = await cluster.scheduler_service()
            assert isinstance(scheduler_service, Service)


@pytest.mark.anyio
async def test_object_dask_worker_group(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_cluster: Callable[..., AsyncContextManager[tuple[str, str]]],
    gen_worker_group: Callable[[str], AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        async with (
            gen_cluster() as (cluster_name, ns),
            gen_worker_group("simpleworkergroup.yaml") as (
                additional_workergroup_name,
                _,
            ),
        ):
            cluster = await DaskCluster.get(cluster_name, namespace=ns)
            additional_workergroup = await DaskWorkerGroup.get(
                additional_workergroup_name, namespace=ns
            )

            worker_groups: list[DaskWorkerGroup] = []
            while not worker_groups:
                worker_groups = await cluster.worker_groups()
                await asyncio.sleep(0.1)
            assert len(worker_groups) == 1  # Just the default worker group
            worker_groups = worker_groups + [additional_workergroup]

            for wg in worker_groups:
                assert isinstance(wg, DaskWorkerGroup)

                deployments: list[Deployment] = []
                while not deployments:
                    deployments = await wg.deployments()
                    await asyncio.sleep(0.1)
                assert all([isinstance(d, Deployment) for d in deployments])

                pods: list[Pod] = []
                while not pods:
                    pods = await wg.pods()
                    await asyncio.sleep(0.1)
                assert all([isinstance(p, Pod) for p in pods])

                assert (await wg.cluster()).name == cluster.name

                for deployment in deployments:
                    assert deployment.labels["dask.org/cluster-name"] == cluster.name
                    for env in deployment.spec["template"]["spec"]["containers"][0][
                        "env"
                    ]:
                        if env["name"] == "DASK_WORKER_NAME":
                            if wg.name == additional_workergroup_name:
                                assert env["value"] == "test-worker"
                            else:
                                assert env["value"] == deployment.name
                        if env["name"] == "DASK_SCHEDULER_ADDRESS":
                            scheduler_service = await cluster.scheduler_service()
                            assert f"{scheduler_service.name}.{ns}" in env["value"]


@pytest.mark.anyio
@pytest.mark.skip(reason="Flaky in CI")
async def test_object_dask_job(
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_job: Callable[[str], AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        async with gen_job("simplejob.yaml") as (job_name, ns):
            job = await DaskJob.get(job_name, namespace=ns)

            job_pod = await job.pod()
            assert isinstance(job_pod, Pod)

            cluster = await job.cluster()
            assert isinstance(cluster, DaskCluster)


async def _get_cluster_status(
    k8s_cluster: KindCluster, ns: str, cluster_name: str
) -> str:
    """
    Will loop infinitely in search of non-falsey cluster status.
    Make sure there is a timeout on any test which calls this.
    """
    while True:
        cluster_status: str = k8s_cluster.kubectl(
            "get",
            "-n",
            ns,
            "daskcluster.kubernetes.dask.org",
            cluster_name,
            "-o",
            "jsonpath='{.status.phase}'",
        ).strip("'")
        if cluster_status:
            return cluster_status
        await asyncio.sleep(0.1)


@pytest.mark.timeout(180)
@pytest.mark.anyio
@pytest.mark.parametrize(
    "cluster_name,expected_status",
    [
        ("valid-name", {"Created", "Pending", "Running"}),
        ((MAX_CLUSTER_NAME_LEN + 1) * "a", {"Error"}),
        ("invalid.chars.in.name", {"Error"}),
    ],
)
async def test_create_cluster_validates_name(
    cluster_name: str,
    expected_status: set[str],
    k8s_cluster: KindCluster,
    kopf_runner: KopfRunner,
    gen_cluster: Callable[..., AsyncContextManager[tuple[str, str]]],
) -> None:
    with kopf_runner:
        async with gen_cluster(cluster_name=cluster_name) as (_, ns):
            actual_status = await _get_cluster_status(k8s_cluster, ns, cluster_name)
            assert actual_status in expected_status


@pytest.mark.anyio
async def test_retire_workers_lifo():
    class Worker:
        def __init__(self, name):
            self.name = name

    workers = [
        Worker(name="worker-1"),
        Worker(name="worker-2"),
        Worker(name="worker-3"),
        Worker(name="worker-4"),
        Worker(name="worker-5"),
    ]

    retired_workers = retire_workers_lifo(workers, 2)

    # Verify we got back exactly 2 workers and they are the last ones
    assert len(retired_workers) == 2
    assert retired_workers == ["worker-4", "worker-5"]



"""Validate resources"""

from dask_kubernetes.constants import MAX_CLUSTER_NAME_LEN, VALID_CLUSTER_NAME
from dask_kubernetes.exceptions import ValidationError


def validate_cluster_name(cluster_name: str) -> None:
    """Raise exception if cluster name is too long and/or has invalid characters"""
    if not VALID_CLUSTER_NAME.match(cluster_name):
        raise ValidationError(
            message=(
                f"The DaskCluster {cluster_name} is invalid: a lowercase RFC 1123 subdomain must "
                "consist of lower case alphanumeric characters, '-' or '.', and must start "
                "and end with an alphanumeric character. DaskCluster name must also be under "
                f"{MAX_CLUSTER_NAME_LEN} characters."
            )
        )



from .kubecluster import (
    KubeCluster,
    make_cluster_spec,
    make_scheduler_spec,
    make_worker_spec,
)
from .discovery import discover

__all__ = [
    "KubeCluster",
    "make_cluster_spec",
    "make_scheduler_spec",
    "make_worker_spec",
    "discover",
]



from . import config

__all__ = []

try:
    from ._version import version as __version__  # noqa
    from ._version import version_tuple as __version_tuple__  # noqa
except ImportError:
    __version__ = "0.0.0"
    __version_tuple__ = (0, 0, 0)



from __future__ import annotations

import asyncio
import copy
import logging
import time
from collections import defaultdict
from contextlib import suppress
from datetime import datetime
from importlib.metadata import entry_points
from typing import TYPE_CHECKING, Any, Final
from uuid import uuid4

import aiohttp
import dask.config
import kopf
import kr8s
from distributed.core import clean_exception, rpc
from distributed.protocol.pickle import dumps
from kr8s.asyncio.objects import Deployment, Pod, Service

from dask_kubernetes.constants import SCHEDULER_NAME_TEMPLATE
from dask_kubernetes.exceptions import ValidationError
from dask_kubernetes.operator._objects import (
    DaskAutoscaler,
    DaskCluster,
    DaskJob,
    DaskWorkerGroup,
)
from dask_kubernetes.operator.networking import get_scheduler_address
from dask_kubernetes.operator.validation import validate_cluster_name

if TYPE_CHECKING:
    from distributed import Scheduler

_ANNOTATION_NAMESPACES_TO_IGNORE: Final[tuple[str, ...]] = (
    "kopf.zalando.org",
    "kubectl.kubernetes.io",
)
_LABEL_NAMESPACES_TO_IGNORE: Final[tuple[str, ...]] = ()

KUBERNETES_DATETIME_FORMAT: Final[str] = "%Y-%m-%dT%H:%M:%SZ"

DASK_AUTOSCALER_COOLDOWN_UNTIL_ANNOTATION: Final[
    str
] = "kubernetes.dask.org/cooldown-until"

# Load operator plugins from other packages
PLUGINS: list[Any] = []
for ep in entry_points(group="dask_operator_plugin"):
    with suppress(AttributeError, ImportError):
        PLUGINS.append(ep.load())


class SchedulerCommError(Exception):
    """Raised when unable to communicate with a scheduler."""


def _get_annotations(meta: kopf.Meta) -> dict[str, str]:
    return {
        annotation_key: annotation_value
        for annotation_key, annotation_value in meta.annotations.items()
        if not any(
            annotation_key.startswith(namespace)
            for namespace in _ANNOTATION_NAMESPACES_TO_IGNORE
        )
    }


def _get_labels(meta: kopf.Meta) -> dict[str, str]:
    return {
        label_key: label_value
        for label_key, label_value in meta.labels.items()
        if not any(
            label_key.startswith(namespace) for namespace in _LABEL_NAMESPACES_TO_IGNORE
        )
    }


def build_scheduler_deployment_spec(
    cluster_name: str,
    pod_spec: kopf.Spec,
    annotations: kopf.Annotations,
    labels: kopf.Labels,
) -> dict[str, Any]:
    labels = dict(labels) | {
        "dask.org/cluster-name": cluster_name,
        "dask.org/component": "scheduler",
        "sidecar.istio.io/inject": "false",
    }
    metadata = {
        "name": SCHEDULER_NAME_TEMPLATE.format(cluster_name=cluster_name),
        "labels": labels,
        "annotations": annotations,
    }
    spec = {
        "replicas": 1,
        "selector": {
            "matchLabels": labels,
        },
        "template": {
            "metadata": metadata,
            "spec": pod_spec,
        },
    }
    return {
        "apiVersion": "apps/v1",
        "kind": "Deployment",
        "metadata": metadata,
        "spec": spec,
    }


def build_scheduler_service_spec(
    cluster_name: str,
    spec: kopf.Spec,
    annotations: kopf.Annotations,
    labels: kopf.Labels,
) -> dict[str, Any]:
    labels = dict(labels) | {
        "dask.org/cluster-name": cluster_name,
        "dask.org/component": "scheduler",
    }
    return {
        "apiVersion": "v1",
        "kind": "Service",
        "metadata": {
            "name": SCHEDULER_NAME_TEMPLATE.format(cluster_name=cluster_name),
            "labels": labels,
            "annotations": annotations,
        },
        "spec": spec,
    }


def build_worker_deployment_spec(
    worker_group_name: str,
    namespace: str,
    cluster_name: str,
    uuid: str,
    pod_spec: kopf.Spec,
    annotations: kopf.Annotations,
    labels: kopf.Labels,
) -> dict[str, Any]:
    labels = dict(labels) | {
        "dask.org/cluster-name": cluster_name,
        "dask.org/workergroup-name": worker_group_name,
        "dask.org/component": "worker",
        "sidecar.istio.io/inject": "false",
    }
    worker_name = f"{worker_group_name}-worker-{uuid}"
    metadata = {
        "name": worker_name,
        "labels": labels,
        "annotations": annotations,
    }
    deployment_spec: dict[str, Any] = {
        "apiVersion": "apps/v1",
        "kind": "Deployment",
        "metadata": metadata,
        "spec": {
            "replicas": 1,
            "selector": {
                "matchLabels": labels,
            },
            "template": {
                "metadata": metadata,
                "spec": copy.deepcopy(pod_spec),
            },
        },
    }
    worker_env = {
        "name": "DASK_WORKER_NAME",
        "value": worker_name,
    }
    scheduler_env = {
        "name": "DASK_SCHEDULER_ADDRESS",
        "value": f"tcp://{cluster_name}-scheduler.{namespace}.svc.cluster.local:8786",
    }
    for container in deployment_spec["spec"]["template"]["spec"]["containers"]:
        if "env" not in container:
            container["env"] = [worker_env, scheduler_env]
            continue

        container_env_names = [env_item["name"] for env_item in container["env"]]

        if "DASK_WORKER_NAME" not in container_env_names:
            container["env"].append(worker_env)
        if "DASK_SCHEDULER_ADDRESS" not in container_env_names:
            container["env"].append(scheduler_env)
    return deployment_spec


def get_job_runner_pod_name(job_name: str) -> str:
    return f"{job_name}-runner"


def build_job_pod_spec(
    job_name: str,
    cluster_name: str,
    namespace: str,
    spec: kopf.Spec,
    annotations: kopf.Annotations,
    labels: kopf.Labels,
) -> dict[str, Any]:
    labels = dict(labels) | {
        "dask.org/cluster-name": cluster_name,
        "dask.org/component": "job-runner",
        "sidecar.istio.io/inject": "false",
    }
    pod_spec: dict[str, Any] = {
        "apiVersion": "v1",
        "kind": "Pod",
        "metadata": {
            "name": get_job_runner_pod_name(job_name),
            "labels": labels,
            "annotations": annotations,
        },
        "spec": copy.deepcopy(spec),
    }
    scheduler_env = {
        "name": "DASK_SCHEDULER_ADDRESS",
        "value": f"tcp://{cluster_name}-scheduler.{namespace}.svc.cluster.local:8786",
    }
    for container in pod_spec["spec"]["containers"]:
        if "env" not in container:
            container["env"] = [scheduler_env]
            continue

        container_env_names = [env_item["name"] for env_item in container["env"]]

        if "DASK_SCHEDULER_ADDRESS" not in container_env_names:
            container["env"].append(scheduler_env)
    return pod_spec


def build_default_worker_group_spec(
    cluster_name: str,
    spec: kopf.Spec,
    annotations: kopf.Annotations,
    labels: kopf.Labels,
) -> dict[str, Any]:
    return {
        "apiVersion": "kubernetes.dask.org/v1",
        "kind": "DaskWorkerGroup",
        "metadata": {
            "name": f"{cluster_name}-default",
            "labels": dict(labels)
            | {
                "dask.org/cluster-name": cluster_name,
                "dask.org/component": "workergroup",
            },
            "annotations": annotations,
        },
        "spec": {
            "cluster": cluster_name,
            "worker": spec,
        },
    }


def build_cluster_spec(
    name: str,
    worker_spec: kopf.Spec,
    scheduler_spec: kopf.Spec,
    annotations: kopf.Annotations,
    labels: kopf.Labels,
) -> dict[str, Any]:
    return {
        "apiVersion": "kubernetes.dask.org/v1",
        "kind": "DaskCluster",
        "metadata": {
            "name": name,
            "labels": dict(labels)
            | {
                "dask.org/cluster-name": name,
            },
            "annotations": annotations,
        },
        "spec": {"worker": worker_spec, "scheduler": scheduler_spec},
    }


@kopf.on.startup()
async def startup(settings: kopf.OperatorSettings, **__: Any) -> None:
    # Set server and client timeouts to reconnect from time to time.
    # In rare occasions the connection might go idle we will no longer receive any events.
    # These timeouts should help in those cases.
    # https://github.com/nolar/kopf/issues/698
    # https://github.com/nolar/kopf/issues/204
    settings.watching.server_timeout = 120
    settings.watching.client_timeout = 150
    settings.watching.connect_timeout = 5

    # The default timeout is 300s which is usually to long
    # https://kopf.readthedocs.io/en/latest/configuration/#networking-timeouts
    settings.networking.request_timeout = 10

    # Configure events
    event_enabled = dask.config.get("kubernetes.controller.events.enabled", True)
    settings.posting.enabled = event_enabled
    event_level = dask.config.get("kubernetes.controller.events.level", "INFO")
    if isinstance(event_level, str) and hasattr(logging, event_level):
        event_level = getattr(logging, event_level)
    if isinstance(event_level, int):
        settings.posting.level = event_level


# There may be useful things for us to expose via the liveness probe
# https://kopf.readthedocs.io/en/stable/probing/#probe-handlers
@kopf.on.probe(id="now")
def get_current_timestamp(**__: Any) -> str:
    return datetime.utcnow().isoformat()


@kopf.on.create("daskcluster.kubernetes.dask.org")
async def daskcluster_create(
    name: str | None,
    namespace: str | None,
    patch: kopf.Patch,
    logger: kopf.Logger,
    **__: Any,
) -> None:
    """When DaskCluster resource is created set the status.phase.

    This allows us to track that the operator is running.
    """
    assert name
    assert namespace
    logger.info(f"DaskCluster {name} created in {namespace}.")
    try:
        validate_cluster_name(name)
    except ValidationError as validation_exc:
        patch.status["phase"] = "Error"
        raise kopf.PermanentError(validation_exc.message)

    patch.status["phase"] = "Created"


@kopf.on.field("daskcluster.kubernetes.dask.org", field="status.phase", new="Created")
async def daskcluster_create_components(
    spec: kopf.Spec,
    name: str | None,
    namespace: str | None,
    logger: kopf.Logger,
    patch: kopf.Patch,
    meta: kopf.Meta,
    **__: Any,
) -> None:
    """When the DaskCluster status.phase goes into Created create the cluster components."""
    assert name
    assert namespace
    logger.info("Creating Dask cluster components.")

    # Create scheduler deployment
    annotations = _get_annotations(meta)
    labels = _get_labels(meta)
    scheduler_spec = spec.get("scheduler", {})
    if "metadata" in scheduler_spec:
        if "annotations" in scheduler_spec["metadata"]:
            annotations.update(**scheduler_spec["metadata"]["annotations"])
        if "labels" in scheduler_spec["metadata"]:
            labels.update(**scheduler_spec["metadata"]["labels"])
    data = build_scheduler_deployment_spec(
        name, scheduler_spec.get("spec"), annotations, labels
    )
    kopf.adopt(data)
    scheduler_deployment = await Deployment(data, namespace=namespace)
    if not await scheduler_deployment.exists():
        await scheduler_deployment.create()
    logger.info(
        f"Scheduler deployment {scheduler_deployment.name} created in {namespace}."
    )

    # Create scheduler service
    data = build_scheduler_service_spec(
        name, scheduler_spec.get("service"), annotations, labels
    )
    kopf.adopt(data)
    scheduler_service = await Service(data, namespace=namespace)
    if not await scheduler_service.exists():
        await scheduler_service.create()
    logger.info(f"Scheduler service {data['metadata']['name']} created in {namespace}.")

    # Create default worker group
    worker_spec = spec.get("worker", {})
    annotations = _get_annotations(meta)
    labels = _get_labels(meta)
    if "metadata" in worker_spec:
        if "annotations" in worker_spec["metadata"]:
            annotations.update(**worker_spec["metadata"]["annotations"])
        if "labels" in worker_spec["metadata"]:
            labels.update(**worker_spec["metadata"]["labels"])
    data = build_default_worker_group_spec(name, worker_spec, annotations, labels)
    worker_group = await DaskWorkerGroup(data, namespace=namespace)
    if not await worker_group.exists():
        await worker_group.create()
    logger.info(f"Worker group {data['metadata']['name']} created in {namespace}.")

    patch.status["phase"] = "Pending"


@kopf.on.field("service", field="status", labels={"dask.org/component": "scheduler"})
async def handle_scheduler_service_status(
    spec: kopf.Spec,
    labels: kopf.Labels,
    status: kopf.Status,
    namespace: str | None,
    **__: Any,
) -> None:
    assert namespace
    # If the Service is a LoadBalancer with no ingress endpoints mark the cluster as Pending
    if spec["type"] == "LoadBalancer" and not len(
        status.get("loadBalancer", {}).get("ingress", [])
    ):
        phase = "Pending"
    # Otherwise mark it as Running
    else:
        phase = "Running"
    cluster = await DaskCluster.get(
        labels["dask.org/cluster-name"], namespace=namespace
    )
    await cluster.patch({"status": {"phase": phase}})


@kopf.on.create("daskworkergroup.kubernetes.dask.org")
async def daskworkergroup_create(
    body: kopf.Body, namespace: str | None, logger: kopf.Logger, **kwargs: Any
) -> None:
    assert namespace
    wg = await DaskWorkerGroup(body, namespace=namespace)
    cluster = await wg.cluster()
    await cluster.adopt(wg)
    logger.info(f"Successfully adopted by {cluster.name}")

    del kwargs["new"]
    await daskworkergroup_replica_update(  # type: ignore[misc]
        body=body,
        logger=logger,
        new=wg.replicas,
        namespace=namespace,
        **kwargs,
    )


async def retire_workers(
    n_workers: int,
    scheduler_service_name: str,
    worker_group_name: str,
    namespace: str | None,
    logger: kopf.Logger,
) -> list[str]:
    assert namespace
    # Try gracefully retiring via the HTTP API
    dashboard_address = await get_scheduler_address(
        scheduler_service_name,
        namespace,
        port_name="http-dashboard",
        allow_external=False,
    )
    async with aiohttp.ClientSession() as session:
        url = f"{dashboard_address}/api/v1/retire_workers"
        params = {"n": n_workers}
        async with session.post(url, json=params) as resp:
            if resp.status <= 300:
                retired_workers = await resp.json()
                logger.info("Retired workers %s", retired_workers)
                return [retired_workers[w]["name"] for w in retired_workers.keys()]
            logger.debug(
                "Received %d response from scheduler API with body %s",
                resp.status,
                await resp.text(),
            )

    # Otherwise try gracefully retiring via the RPC
    logger.debug(
        f"Scaling {worker_group_name} failed via the HTTP API, falling back to the Dask RPC"
    )
    # Dask version mismatches between the operator and scheduler may cause this to fail in any number of unexpected ways
    with suppress(Exception):
        comm_address = await get_scheduler_address(
            scheduler_service_name,
            namespace,
            allow_external=False,
        )
        async with rpc(comm_address) as scheduler_comm:
            workers_to_close = await scheduler_comm.workers_to_close(
                n=n_workers,
                attribute="name",
            )
            await scheduler_comm.retire_workers(names=workers_to_close)
            if isinstance(workers_to_close, tuple):
                workers_to_close = list(workers_to_close)
            assert isinstance(workers_to_close, list)
            return workers_to_close

    # Finally fall back to last-in-first-out scaling
    logger.warning(
        f"Scaling {worker_group_name} failed via the HTTP API and the Dask RPC, falling back to LIFO scaling. "
        "This can result in lost data, see https://kubernetes.dask.org/en/latest/operator_troubleshooting.html."
    )
    workers = [
        deployment
        async for deployment in kr8s.asyncio.get(
            "deployments",
            namespace=namespace,
            label_selector={"dask.org/workergroup-name": worker_group_name},
        )
    ]
    return retire_workers_lifo(workers, n_workers)


def retire_workers_lifo(workers, n_workers: int) -> list[str]:
    return [w.name for w in workers[-n_workers:]]


async def check_scheduler_idle(
    scheduler_service_name: str, namespace: str | None, logger: kopf.Logger
) -> float:
    assert namespace
    # Try getting idle time via HTTP API
    dashboard_address = await get_scheduler_address(
        scheduler_service_name,
        namespace,
        port_name="http-dashboard",
        allow_external=False,
    )
    async with aiohttp.ClientSession() as session:
        url = f"{dashboard_address}/api/v1/check_idle"
        async with session.get(url) as resp:
            if resp.status <= 300:
                idle_since = (await resp.json())["idle_since"]
                if idle_since:
                    logger.debug("Scheduler idle since: %s", idle_since)
                return float(idle_since)
            logger.debug(
                "Received %d response from scheduler API with body %s",
                resp.status,
                await resp.text(),
            )

    # Otherwise try gracefully checking via the RPC
    logger.debug(
        f"Checking {scheduler_service_name} idleness failed via the HTTP API, falling back to the Dask RPC"
    )
    # Dask version mismatches between the operator and scheduler may cause this to fail in any number of unexpected ways
    with suppress(Exception):
        comm_address = await get_scheduler_address(
            scheduler_service_name,
            namespace,
            allow_external=False,
        )
        async with rpc(comm_address) as scheduler_comm:
            idle_since = await scheduler_comm.check_idle()
            if idle_since:
                logger.debug("Scheduler idle since: %s", idle_since)
            return float(idle_since)

    # Finally fall back to code injection via the Dask RPC for distributed<=2023.3.1
    logger.debug(
        f"Checking {scheduler_service_name} idleness failed via the Dask RPC, falling back to run_on_scheduler"
    )

    def idle_since_func(dask_scheduler: Scheduler) -> float:
        if not dask_scheduler.idle_timeout:
            dask_scheduler.idle_timeout = 300
        dask_scheduler.check_idle()
        assert dask_scheduler.idle_since
        return dask_scheduler.idle_since

    comm_address = await get_scheduler_address(
        scheduler_service_name,
        namespace,
        allow_external=False,
    )
    async with rpc(comm_address) as scheduler_comm:
        response = await scheduler_comm.run_function(
            function=dumps(idle_since_func),
        )
        if response["status"] == "error":
            typ, exc, tb = clean_exception(**response)
            assert exc
            raise exc.with_traceback(tb)
        else:
            idle_since = response["result"]
            if idle_since:
                logger.debug("Scheduler idle since: %s", idle_since)
            return float(idle_since)


async def get_desired_workers(
    scheduler_service_name: str, namespace: str | None
) -> Any:
    assert namespace
    # Try gracefully retiring via the HTTP API
    dashboard_address = await get_scheduler_address(
        scheduler_service_name,
        namespace,
        port_name="http-dashboard",
        allow_external=False,
    )
    async with aiohttp.ClientSession() as session:
        url = f"{dashboard_address}/api/v1/adaptive_target"
        async with session.get(url) as resp:
            if resp.status <= 300:
                desired_workers = await resp.json()
                return desired_workers["workers"]

    # Otherwise try gracefully retiring via the RPC
    # Dask version mismatches between the operator and scheduler may cause this to fail in any number of unexpected ways
    try:
        comm_address = await get_scheduler_address(
            scheduler_service_name,
            namespace,
            allow_external=False,
        )
        async with rpc(comm_address) as scheduler_comm:
            return await scheduler_comm.adaptive_target()
    except Exception as e:
        raise SchedulerCommError(
            "Unable to get number of desired workers from scheduler"
        ) from e


worker_group_scale_locks: dict[str, asyncio.Lock] = defaultdict(lambda: asyncio.Lock())


@kopf.on.field("daskcluster.kubernetes.dask.org", field="spec.worker.replicas")
async def daskcluster_default_worker_group_replica_update(
    name: str | None,
    namespace: str | None,
    old: Any | None,
    new: Any | None,
    **__: Any,
) -> None:
    assert name
    assert namespace
    if old is not None:
        wg = await DaskWorkerGroup.get(f"{name}-default", namespace=namespace)
        assert isinstance(new, int)
        await wg.scale(new)


@kopf.on.field("daskworkergroup.kubernetes.dask.org", field="spec.worker.replicas")
async def daskworkergroup_replica_update(
    name: str | None,
    namespace: str | None,
    meta: kopf.Meta,
    spec: kopf.Spec,
    new: Any | None,
    body: kopf.Body,
    logger: kopf.Logger,
    **__: Any,
) -> None:
    assert name
    assert namespace
    cluster_name = spec["cluster"]
    wg = await DaskWorkerGroup(body, namespace=namespace)
    try:
        cluster = await wg.cluster()
    except kr8s.NotFoundError:
        # No need to scale if cluster is deleted, pods will be cleaned up
        return

    # Replica updates can come in quick succession and the changes must be applied atomically to ensure
    # the number of workers ends in the correct state
    async with worker_group_scale_locks[f"{namespace}/{name}"]:
        current_workers = len(
            [
                deployment
                async for deployment in kr8s.asyncio.get(
                    "deployments",
                    namespace=namespace,
                    label_selector={"dask.org/workergroup-name": name},
                )
            ]
        )
        assert isinstance(new, int)
        desired_workers = new
        workers_needed = desired_workers - current_workers
        labels = _get_labels(meta)
        annotations = _get_annotations(meta)
        worker_spec = spec["worker"]
        if "metadata" in worker_spec:
            if "annotations" in worker_spec["metadata"]:
                annotations.update(**worker_spec["metadata"]["annotations"])
            if "labels" in worker_spec["metadata"]:
                labels.update(**worker_spec["metadata"]["labels"])

        batch_size = int(
            dask.config.get("kubernetes.controller.worker-allocation.batch-size") or 0
        )
        batch_size = min(workers_needed, batch_size) if batch_size else workers_needed
        batch_delay = int(
            dask.config.get("kubernetes.controller.worker-allocation.delay") or 0
        )
        if workers_needed > 0:
            for _ in range(batch_size):
                data = build_worker_deployment_spec(
                    worker_group_name=name,
                    namespace=namespace,
                    cluster_name=cluster_name,
                    uuid=uuid4().hex[:10],
                    pod_spec=worker_spec["spec"],
                    annotations=annotations,
                    labels=labels,
                )
                kopf.adopt(data, owner=body)
                kopf.label(data, labels=cluster.labels)
                worker_deployment = await Deployment(data, namespace=namespace)
                await worker_deployment.create()
        if workers_needed > batch_size:
            raise kopf.TemporaryError(
                "Added maximum number of workers for this batch but still need to create more workers, "
                f"waiting for {batch_delay} seconds before continuing.",
                delay=batch_delay,
            )
        logger.info(f"Scaled worker group {name} up to {desired_workers} workers.")
        if workers_needed < 0:
            worker_ids = await retire_workers(
                n_workers=-workers_needed,
                scheduler_service_name=SCHEDULER_NAME_TEMPLATE.format(
                    cluster_name=cluster_name
                ),
                worker_group_name=name,
                namespace=namespace,
                logger=logger,
            )
            logger.info(f"Workers to close: {worker_ids}")
            for wid in worker_ids:
                worker_deployment = await Deployment(wid, namespace=namespace)
                await worker_deployment.delete()
            logger.info(
                f"Scaled worker group {name} down to {desired_workers} workers."
            )


@kopf.on.delete("daskworkergroup.kubernetes.dask.org", optional=True)
async def daskworkergroup_remove(
    name: str | None, namespace: str | None, **__: Any
) -> None:
    assert name
    assert namespace
    lock_key = f"{name}/{namespace}"
    if lock_key in worker_group_scale_locks:
        del worker_group_scale_locks[lock_key]


@kopf.on.create("daskjob.kubernetes.dask.org")
async def daskjob_create(
    name: str | None,
    namespace: str | None,
    logger: kopf.Logger,
    patch: kopf.Patch,
    **__: Any,
) -> None:
    assert name
    assert namespace
    logger.info(f"A DaskJob has been created called {name} in {namespace}.")
    patch.status["jobStatus"] = "JobCreated"


@kopf.on.field(
    "daskjob.kubernetes.dask.org", field="status.jobStatus", new="JobCreated"
)
async def daskjob_create_components(
    spec: kopf.Spec,
    name: str | None,
    namespace: str | None,
    logger: kopf.Logger,
    patch: kopf.Patch,
    meta: kopf.Meta,
    **__: Any,
) -> None:
    assert name
    assert namespace
    logger.info("Creating Dask job components.")
    cluster_name = f"{name}"
    labels = _get_labels(meta)
    annotations = _get_annotations(meta)
    cluster_spec = spec["cluster"]
    if "metadata" in cluster_spec:
        if "annotations" in cluster_spec["metadata"]:
            annotations.update(**cluster_spec["metadata"]["annotations"])
        if "labels" in cluster_spec["metadata"]:
            labels.update(**cluster_spec["metadata"]["labels"])
    cluster_spec = build_cluster_spec(
        cluster_name,
        cluster_spec["spec"]["worker"],
        cluster_spec["spec"]["scheduler"],
        annotations,
        labels,
    )
    kopf.adopt(cluster_spec)
    cluster = await DaskCluster(cluster_spec, namespace=namespace)
    if not await cluster.exists():
        await cluster.create()
        logger.info(
            f"Cluster {cluster_spec['metadata']['name']} for job {name} created in {namespace}."
        )

    labels = _get_labels(meta)
    annotations = _get_annotations(meta)
    job_spec = spec["job"]
    if "metadata" in job_spec:
        if "annotations" in job_spec["metadata"]:
            annotations.update(**job_spec["metadata"]["annotations"])
        if "labels" in job_spec["metadata"]:
            labels.update(**job_spec["metadata"]["labels"])
    job_pod_spec = build_job_pod_spec(
        job_name=name,
        cluster_name=cluster_name,
        namespace=namespace,
        spec=job_spec["spec"],
        annotations=annotations,
        labels=labels,
    )
    kopf.adopt(job_pod_spec)
    job_pod = await Pod(job_pod_spec, namespace=namespace)
    if not await job_pod.exists():
        await job_pod.create()
    patch.status["clusterName"] = cluster_name
    patch.status["jobStatus"] = "ClusterCreated"
    patch.status["jobRunnerPodName"] = get_job_runner_pod_name(name)


@kopf.on.field(
    "pod",
    field="status.phase",
    labels={"dask.org/component": "job-runner"},
    new="Running",
)
async def handle_runner_status_change_running(
    meta: kopf.Meta, namespace: str | None, logger: kopf.Logger, **__: Any
) -> None:
    assert namespace
    logger.info("Job now in running")
    name = meta["labels"]["dask.org/cluster-name"]
    job = await DaskJob.get(name, namespace=namespace)
    await job.patch(
        {
            "status": {
                "jobStatus": "Running",
                "startTime": datetime.utcnow().strftime(KUBERNETES_DATETIME_FORMAT),
            }
        },
        subresource="status",
    )


@kopf.on.field(
    "pod",
    field="status.phase",
    labels={"dask.org/component": "job-runner"},
    new="Succeeded",
)
async def handle_runner_status_change_succeeded(
    meta: kopf.Meta, namespace: str | None, logger: kopf.Logger, **__: Any
) -> None:
    assert namespace
    logger.info("Job succeeded, deleting Dask cluster.")
    name = meta["labels"]["dask.org/cluster-name"]
    cluster = await DaskCluster.get(name, namespace=namespace)
    await cluster.delete()
    job = await DaskJob.get(name, namespace=namespace)
    await job.patch(
        {
            "status": {
                "jobStatus": "Successful",
                "endTime": datetime.utcnow().strftime(KUBERNETES_DATETIME_FORMAT),
            }
        },
        subresource="status",
    )


@kopf.on.field(
    "pod",
    field="status.phase",
    labels={"dask.org/component": "job-runner"},
    new="Failed",
)
async def handle_runner_status_change_failed(
    meta: kopf.Meta, namespace: str | None, logger: kopf.Logger, **__: Any
) -> None:
    assert namespace
    logger.info("Job failed, deleting Dask cluster.")
    name = meta["labels"]["dask.org/cluster-name"]
    cluster = await DaskCluster.get(name, namespace=namespace)
    await cluster.delete()
    job = await DaskJob.get(name, namespace=namespace)
    await job.patch(
        {
            "status": {
                "jobStatus": "Failed",
                "endTime": datetime.utcnow().strftime(KUBERNETES_DATETIME_FORMAT),
            }
        },
        subresource="status",
    )


@kopf.on.create("daskautoscaler.kubernetes.dask.org")
async def daskautoscaler_create(
    body: kopf.Body, logger: kopf.Logger, **__: Any
) -> None:
    """When an autoscaler is created make it a child of the associated cluster for cascade deletion."""
    autoscaler = await DaskAutoscaler(body)
    cluster = await autoscaler.cluster()
    await cluster.adopt(autoscaler)
    logger.info(f"Autoscaler {autoscaler.name} adopted by cluster {cluster.name}")


@kopf.timer("daskautoscaler.kubernetes.dask.org", interval=5.0)
async def daskautoscaler_adapt(
    spec: kopf.Spec,
    name: str | None,
    namespace: str | None,
    logger: kopf.Logger,
    **__: Any,
) -> None:
    assert name
    assert namespace
    try:
        scheduler = await Pod.get(
            label_selector={
                "dask.org/component": "scheduler",
                "dask.org/cluster-name": spec["cluster"],
            },
            namespace=namespace,
        )
        if not await scheduler.ready():
            raise ValueError
    except ValueError:
        logger.info("Scheduler not ready, skipping autoscaling")
        return

    autoscaler = await DaskAutoscaler.get(name, namespace=namespace)
    worker_group = await DaskWorkerGroup.get(
        f"{spec['cluster']}-default", namespace=namespace
    )

    current_replicas = worker_group.replicas
    cooldown_until = float(
        autoscaler.annotations.get(
            DASK_AUTOSCALER_COOLDOWN_UNTIL_ANNOTATION, time.time()
        )
    )

    # Cooldown autoscaling to prevent thrashing
    if time.time() < cooldown_until:
        logger.debug("Autoscaler for %s is in cooldown", spec["cluster"])
        return

    # Ask the scheduler for the desired number of worker
    try:
        desired_workers = await get_desired_workers(
            scheduler_service_name=f"{spec['cluster']}-scheduler",
            namespace=namespace,
        )
    except SchedulerCommError:
        logger.error("Unable to get desired number of workers from scheduler.")
        return

    # Ensure the desired number is within the min and max
    desired_workers = max(spec["minimum"], desired_workers)
    desired_workers = min(spec["maximum"], desired_workers)

    if current_replicas > 0:
        max_scale_down = int(current_replicas * 0.25)
        max_scale_down = 1 if max_scale_down == 0 else max_scale_down
        desired_workers = max(current_replicas - max_scale_down, desired_workers)

    # Update the default DaskWorkerGroup
    if desired_workers != current_replicas:
        await worker_group.scale(desired_workers)

        cooldown_until = time.time() + 15

        await autoscaler.annotate(
            {DASK_AUTOSCALER_COOLDOWN_UNTIL_ANNOTATION: str(cooldown_until)}
        )

        logger.info(
            "Autoscaler updated %s worker count from %d to %d",
            spec["cluster"],
            current_replicas,
            desired_workers,
        )
    else:
        logger.debug(
            "Not autoscaling %s with %d workers", spec["cluster"], current_replicas
        )


@kopf.timer("daskcluster.kubernetes.dask.org", interval=5.0)
async def daskcluster_autoshutdown(
    spec: kopf.Spec,
    name: str | None,
    namespace: str | None,
    logger: kopf.Logger,
    **__: Any,
) -> None:
    idle_timeout = spec.get("idleTimeout", 0)
    if idle_timeout:
        try:
            idle_since = await check_scheduler_idle(
                scheduler_service_name=f"{name}-scheduler",
                namespace=namespace,
                logger=logger,
            )
        except Exception:  # TODO: Not use broad "Exception" catch here
            logger.warning(
                "Unable to connect to scheduler, skipping autoshutdown check."
            )
            return
        if idle_since and time.time() > idle_since + idle_timeout:
            cluster = await DaskCluster.get(name, namespace=namespace)
            await cluster.delete()



import logging
import os
import pathlib
import shutil
import subprocess
import sys
import tempfile
import uuid
from typing import Final, Iterator

import kr8s
import pytest
from kopf.testing import KopfRunner
from packaging.version import Version
from pytest_kind.cluster import KindCluster

DIR: Final[pathlib.Path] = pathlib.Path(__file__).parent.absolute()


def check_dependency(dependency):
    if shutil.which(dependency) is None:
        raise RuntimeError(
            f"Missing dependency {dependency}. "
            f"Please install {dependency} following the instructions for your OS. "
        )


check_dependency("helm")
check_dependency("kubectl")
check_dependency("docker")

DISABLE_LOGGERS: Final[list[str]] = ["httpcore", "httpx"]


def pytest_configure() -> None:
    for logger_name in DISABLE_LOGGERS:
        logger = logging.getLogger(logger_name)
        logger.disabled = True


@pytest.fixture()
def kopf_runner(k8s_cluster: KindCluster, namespace: str) -> KopfRunner:
    yield KopfRunner(
        ["run", "-m", "dask_kubernetes.operator", "--verbose", "--namespace", namespace]
    )


@pytest.fixture(scope="session")
def docker_image() -> str:
    image_name = "dask-kubernetes:dev"
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
    subprocess.run(
        [
            "docker",
            "build",
            "-t",
            image_name,
            "--build-arg",
            f"PYTHON={python_version}",
            "./ci/",
        ],
        check=True,
    )
    return image_name


@pytest.fixture(scope="session")
def k8s_cluster(
    request: pytest.FixtureRequest, docker_image: str
) -> Iterator[KindCluster]:
    image = None
    if version := os.environ.get("KUBERNETES_VERSION"):
        image = f"kindest/node:v{version}"

    kind_cluster = KindCluster(
        name="pytest-kind",
        image=image,
    )
    kind_cluster.create()
    os.environ["KUBECONFIG"] = str(kind_cluster.kubeconfig_path)
    kind_cluster.load_docker_image(docker_image)
    yield kind_cluster
    del os.environ["KUBECONFIG"]
    if not request.config.getoption("keep_cluster"):
        kind_cluster.delete()


@pytest.fixture(scope="session", autouse=True)
def install_istio(k8s_cluster: KindCluster) -> None:
    if bool(os.environ.get("TEST_ISTIO", False)):
        check_dependency("istioctl")
        subprocess.run(
            ["istioctl", "install", "--set", "profile=demo", "-y"], check=True
        )
        k8s_cluster.kubectl(
            "label", "namespace", "default", "istio-injection=enabled", "--overwrite"
        )


@pytest.fixture(autouse=True)
def namespace(k8s_cluster: KindCluster) -> Iterator[str]:
    ns = "dask-k8s-pytest-" + uuid.uuid4().hex[:10]
    k8s_cluster.kubectl("create", "ns", ns)
    yield ns
    k8s_cluster.kubectl("delete", "ns", ns, "--wait=false", "--ignore-not-found=true")


@pytest.fixture(scope="session", autouse=True)
def install_gateway(k8s_cluster: KindCluster) -> Iterator[None]:
    kube_version = Version(kr8s.version().get("gitVersion"))
    if bool(os.environ.get("TEST_DASK_GATEWAY", False)):
        check_dependency("helm")
        # To ensure the operator can coexist with Gateway
        subprocess.run(
            [
                "helm",
                "upgrade",
                "dask-gateway",
                "dask-gateway",
                "--install",
                "--version",
                "2024.1.0" if kube_version < Version("1.30.0") else ">=2024.1.0",
                "--repo=https://helm.dask.org",
                "--create-namespace",
                "--namespace",
                "dask-gateway",
            ],
            check=True,
            env={**os.environ},
        )
        yield
        subprocess.run(
            [
                "helm",
                "delete",
                "--namespace",
                "dask-gateway",
                "dask-gateway",
            ],
            check=True,
            env={**os.environ},
        )
    else:
        yield


@pytest.fixture(scope="session", autouse=True)
def customresources(k8s_cluster: KindCluster) -> Iterator[None]:
    temp_dir = tempfile.TemporaryDirectory()
    crd_path = os.path.join(DIR, "operator", "customresources")

    def run_generate(crd_path: str, patch_path: str, temp_path: str) -> None:
        subprocess.run(
            ["k8s-crd-resolver", "-r", "-j", patch_path, crd_path, temp_path],
            check=True,
            env={**os.environ},
        )

    for crd in ["daskcluster", "daskworkergroup", "daskjob", "daskautoscaler"]:
        run_generate(
            os.path.join(crd_path, f"{crd}.yaml"),
            os.path.join(crd_path, f"{crd}.patch.yaml"),
            os.path.join(temp_dir.name, f"{crd}.yaml"),
        )

    k8s_cluster.kubectl("apply", "-f", temp_dir.name)
    yield
    k8s_cluster.kubectl("delete", "--wait=false", "-f", temp_dir.name)
    temp_dir.cleanup()


@pytest.fixture
def anyio_backend() -> str:
    return "asyncio"
